"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const turfHelpers = __importStar(require("@turf/helpers"));
const tiles_1 = require("./tiles");
const fs_1 = require("fs");
const tile_index_1 = require("./tile_index");
const index_1 = require("./index");
const child_process_1 = require("child_process");
const point_matcher_1 = require("./point_matcher");
const length_1 = __importDefault(require("@turf/length"));
const distance_1 = __importDefault(require("@turf/distance"));
const { exec } = require('child_process');
const OSRM = require("osrm");
const xml = require('xml');
const stream = require('stream');
const levelup = require('levelup');
const leveldown = require('leveldown');
const chalk = require('chalk');
const path = require('path');
const util = require('util');
const util_1 = require("./util");
const uuidHash = require('uuid-by-string');
const DEFAULT_SEARCH_RADIUS = 10;
const MIN_CONFIDENCE = 0.5;
const OPTIMIZE_GRAPH = true;
const USE_LOCAL_CACHE = true;
const SHST_GRAPH_CACHE_DIR = 'shst/cache/graphs/';
var MatchType;
(function (MatchType) {
    MatchType["DIRECT"] = "direct";
    MatchType["HMM"] = "hmm";
})(MatchType = exports.MatchType || (exports.MatchType = {}));
var GraphMode;
(function (GraphMode) {
    GraphMode["CAR"] = "car";
    GraphMode["BIKE"] = "bike";
    GraphMode["PEDESTRIAN"] = "ped";
})(GraphMode = exports.GraphMode || (exports.GraphMode = {}));
// interface typecheck for SharedStreetsGeometry
function geomInstance(object) {
    return 'forwardReferenceId' in object;
}
function getReferenceLength(ref) {
    var length = 0;
    for (var lr of ref.locationReferences) {
        if (lr.distanceToNextRef)
            length += lr.distanceToNextRef;
    }
    return length / 100;
}
class GraphNode {
}
class GraphEdge {
}
class GraphNodeEdgeRelations {
}
class LevelDB {
    constructor(directory) {
        this.db = levelup(leveldown(directory));
    }
    get(key) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                var data = yield this.db.get(key);
                return data.toString();
            }
            catch (error) {
                return null;
            }
        });
    }
    put(key, data) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.db.put(key, data);
        });
    }
    has(key) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.db.get(key);
                return true;
            }
            catch (error) {
                return null;
            }
        });
    }
}
exports.LevelDB = LevelDB;
function streamPromise(stream) {
    return new Promise((resolve, reject) => {
        stream.on('end', () => {
            resolve('end');
        });
        stream.on('error', (error) => {
            reject(error);
        });
    });
}
class Graph {
    constructor(extent, params, existingTileIndex = null, graphMode = GraphMode.CAR) {
        // options
        this.searchRadius = DEFAULT_SEARCH_RADIUS;
        this.snapIntersections = true;
        this.useHMM = true;
        this.useDirect = true;
        this.tilePathGroup = tiles_1.TilePathGroup.fromPolygon(extent, 1000, params);
        this.tilePathGroup.addType(tiles_1.TileType.GEOMETRY);
        this.tilePathGroup.addType(tiles_1.TileType.REFERENCE);
        this.graphMode = graphMode;
        var paths = [];
        for (var path of this.tilePathGroup) {
            paths.push(path.toPathString());
        }
        if (existingTileIndex) {
            this.tileIndex = existingTileIndex;
        }
        else {
            this.tileIndex = new tile_index_1.TileIndex();
        }
        this.pointMatcher = new point_matcher_1.PointMatcher(extent, params, this.tileIndex);
        // create id from tile path hash  
        this.id = uuidHash(this.graphMode + ' ' + paths.join(" "));
    }
    createGraphXml() {
        return __awaiter(this, void 0, void 0, function* () {
            // build xml representation of graph + leveldb id map
            var graphPath = path.join(SHST_GRAPH_CACHE_DIR, this.id);
            var nextNodeId = 1;
            var nextEdgeId = 1;
            var xmlPath = path.join(graphPath, '/graph.xml');
            // create xml stream
            const pipeline = util.promisify(stream.pipeline);
            var xmlStreamWriter = fs_1.createWriteStream(xmlPath);
            var osmRootElem = xml.element({ _attr: { version: '0.6', generator: 'shst cli v1.0' } });
            var xmlStream = xml({ osm: osmRootElem }, { stream: true });
            xmlStream.on('data', function (chunk) { xmlStreamWriter.write(chunk); });
            const writeNode = (lon, lat, shstIntersectionId = null) => __awaiter(this, void 0, void 0, function* () {
                var nodeId;
                // check if intersection node already written
                if (shstIntersectionId && (yield this.db.has(shstIntersectionId))) {
                    var node = JSON.parse(yield this.db.get(shstIntersectionId));
                    nodeId = node.nodeId;
                }
                else {
                    nodeId = nextNodeId;
                    var newNode = new GraphNode();
                    newNode.nodeId = nodeId;
                    if (shstIntersectionId) {
                        newNode.shstIntersectionId = shstIntersectionId;
                        yield this.db.put(shstIntersectionId, JSON.stringify(newNode));
                    }
                    yield this.db.put('node:' + nodeId, JSON.stringify(newNode));
                    // write node xml
                    osmRootElem.push({ node: [{ _attr: { id: nodeId, lat: lat, lon: lon } }] });
                    nextNodeId++;
                }
                return nodeId;
            });
            for (var obj of this.tileIndex.objectIndex.values()) {
                if (geomInstance(obj)) {
                    // iterate through coordinates and build nodes
                    var coords = index_1.lonlatsToCoords(obj.lonlats);
                    var nodeIds = [];
                    for (var i = 0; i < coords.length; i++) {
                        var shstIntersectionId = null;
                        if (i === 0)
                            shstIntersectionId = obj.fromIntersectionId;
                        else if (i === coords.length - 1)
                            shstIntersectionId = obj.toIntersectionId;
                        var nodeId = yield writeNode(coords[i][0], coords[i][1], shstIntersectionId);
                        nodeIds.push(nodeId);
                    }
                    var edge = new GraphEdge();
                    edge.shstGeometryId = obj.id;
                    edge.edgeId = nextEdgeId;
                    yield this.db.put(obj.id, JSON.stringify(edge));
                    yield this.db.put('edge:' + edge.edgeId, JSON.stringify(edge));
                    var nodeIdElems = [];
                    for (nodeId of nodeIds) {
                        var nodeEdgeRelation;
                        if (yield this.db.has('node-edges:' + nodeId)) {
                            // insert new edge into existing node-edge relation set for node id
                            nodeEdgeRelation = JSON.parse(yield this.db.get('node-edges:' + nodeId));
                            var edgeIds = new Set(nodeEdgeRelation.edgeIds);
                            edgeIds.add(edge.edgeId);
                            nodeEdgeRelation.edgeIds = [...edgeIds.values()];
                        }
                        else {
                            // create node-edge relation set for node id
                            nodeEdgeRelation = new GraphNodeEdgeRelations();
                            nodeEdgeRelation.nodeId = nodeId;
                            nodeEdgeRelation.edgeIds = [edge.edgeId];
                        }
                        yield this.db.put('node-edges:' + nodeId, JSON.stringify(nodeEdgeRelation));
                        nodeIdElems.push({ nd: [{ _attr: { ref: nodeId } }] });
                    }
                    var oneWay = obj.backReferenceId ? 'no' : 'yes';
                    var roadClass = obj.roadClass.toLocaleLowerCase();
                    osmRootElem.push({ way: [{ _attr: { id: edge.edgeId } }, { tag: { _attr: { k: 'highway', v: roadClass } } }, { tag: { _attr: { k: 'oneway', v: oneWay } } }, ...nodeIdElems] });
                    nextEdgeId++;
                }
            }
            const writeFinished = new Promise((resolve, reject) => {
                xmlStreamWriter.on('finish', () => {
                    resolve(xmlPath);
                });
            });
            osmRootElem.close();
            xmlStreamWriter.close();
            //xmlStream.end();
            return writeFinished;
        });
    }
    buildGraph() {
        return __awaiter(this, void 0, void 0, function* () {
            // check if graph is already built;
            if (this.osrm)
                return;
            try {
                var graphPath = path.join(SHST_GRAPH_CACHE_DIR, this.id);
                var dbPath = path.join(graphPath, '/db');
                yield this.tileIndex.indexTilesByPathGroup(this.tilePathGroup);
                if (USE_LOCAL_CACHE && fs_1.existsSync(dbPath)) {
                    var osrmPath = path.join(graphPath, '/graph.xml.osrm');
                    //console.log(chalk.keyword('lightgreen')("     loading pre-built graph from: " + osrmPath));
                    this.db = new LevelDB(dbPath);
                    if (OPTIMIZE_GRAPH)
                        this.osrm = new OSRM({ path: osrmPath });
                    else
                        this.osrm = new OSRM({ path: osrmPath, algorithm: "MLD" });
                }
                else {
                    // TODO before building, check if this graph is a subset of an existing graph
                    fs_1.mkdirSync(dbPath, { recursive: true });
                    this.db = new LevelDB(dbPath);
                    console.log(chalk.keyword('lightgreen')("     building graph xml..."));
                    var xmlPath = yield this.createGraphXml();
                    //extract 
                    console.log(chalk.keyword('lightgreen')("     building graph from: " + xmlPath));
                    var profile;
                    if (this.graphMode === GraphMode.CAR)
                        profile = 'node_modules/osrm/profiles/car.lua';
                    else if (this.graphMode === GraphMode.BIKE)
                        profile = 'node_modules/osrm/profiles/bicycle.lua';
                    else if (this.graphMode === GraphMode.PEDESTRIAN)
                        profile = 'node_modules/osrm/profiles/foot.lua';
                    child_process_1.execSync('node_modules/osrm/lib/binding/osrm-extract ' + xmlPath + ' -p ' + profile);
                    var osrmPath = xmlPath + '.osrm';
                    if (OPTIMIZE_GRAPH) {
                        console.log(chalk.keyword('lightgreen')("     optimizing graph..."));
                        child_process_1.execSync('node_modules/osrm/lib/binding/osrm-contract ' + osrmPath);
                        this.osrm = new OSRM({ path: osrmPath });
                    }
                    else {
                        child_process_1.execSync('node_modules/osrm/lib/binding/osrm-partition ' + osrmPath);
                        child_process_1.execSync('node_modules/osrm/lib/binding/osrm-customize ' + osrmPath);
                        console.log(chalk.keyword('lightgreen')("     skipping graph optimization..."));
                        this.osrm = new OSRM({ path: osrmPath, algorithm: "MLD" });
                    }
                }
            }
            catch (e) {
                console.log("Unable to build graph: " + e);
                console.log("Try deleting existing cached graph: " + graphPath);
            }
        });
    }
    matchTrace(feature) {
        return __awaiter(this, void 0, void 0, function* () {
            // fall back to hmm for probabilistic path discovery
            if (!this.osrm)
                throw "Graph not buit. call buildGraph() before running queries.";
            var hmmOptions = {
                coordinates: feature.geometry.coordinates,
                annotations: true,
                geometries: 'geojson'
            };
            try {
                var matches = yield new Promise((resolve, reject) => {
                    this.osrm.match(hmmOptions, function (err, response) {
                        if (err)
                            reject(err);
                        else
                            resolve(response);
                    });
                });
                var graphEdgeSequence = [];
                this.tileIndex.featureIndex.get("geomId");
                if (matches['matchings'] && matches['matchings'].length > 0) {
                    var match = matches['matchings'][0];
                    //console.log(match.confidence)
                    // if(match.confidence < MIN_CONFIDENCE)
                    //     continue;
                    // this is kind of convoluted due to the sparse info returned in the OSRM annotations...
                    // write out sequence of nodes and edges as emitted from walking OSRM-returned nodes
                    // finding the actual posistion and directionality of the OSRM-edge within the ShSt graph 
                    // edge means that we have to snap start/end points in the OSRM geom
                    //console.log(JSON.stringify(match.geometry));
                    var edgeCandidates;
                    var nodes = [];
                    var visitedNodes = new Set();
                    // ooof this is brutual -- need to unpack legs and reduce list... 
                    for (var leg of match['legs']) {
                        //console.log(leg['annotation']['nodes'])
                        for (var n of leg['annotation']['nodes']) {
                            if (!visitedNodes.has(n) || nodes.length == 0)
                                nodes.push(n);
                            visitedNodes.add(n);
                        }
                    }
                    var visitedEdges = new Set();
                    // then group node pairs into unique edges...
                    for (var nodeId of nodes) {
                        if (yield this.db.has('node:' + nodeId)) {
                            var node = JSON.parse(yield this.db.get('node:' + nodeId));
                            var nodeEdges = JSON.parse(yield this.db.get('node-edges:' + nodeId));
                            if (!edgeCandidates) {
                                edgeCandidates = new Set([...nodeEdges.edgeIds]);
                            }
                            else {
                                var newCandiates = new Set([...nodeEdges.edgeIds]);
                                edgeCandidates = new Set([...edgeCandidates].filter(x => newCandiates.has(x)));
                            }
                            if (edgeCandidates.size == 1) {
                                var edgeId = [...edgeCandidates][0];
                                // reduce repeating edgeIds down to one per edge
                                if (graphEdgeSequence.length == 0 || graphEdgeSequence[graphEdgeSequence.length - 1].edgeId !== edgeId) {
                                    var edge = JSON.parse(yield this.db.get('edge:' + edgeId));
                                    graphEdgeSequence.push(edge);
                                }
                                edgeCandidates = new Set([...nodeEdges.edgeIds]);
                            }
                        }
                    }
                }
                var pathCandidate = new PathCandidate();
                pathCandidate.matchType = MatchType.HMM;
                pathCandidate.score = match.confidence;
                pathCandidate.originalFeature = feature;
                pathCandidate.segments = [];
                var segmentGeoms = turfHelpers.multiLineString([]);
                for (var geom of graphEdgeSequence) {
                    var pathSegment = new PathSegment();
                    pathSegment.geometryId = geom.shstGeometryId;
                    // TODO calc directionality from graph edge trajectory possible...
                    pathCandidate.segments.push(pathSegment);
                    var segementGeom = this.tileIndex.featureIndex.get(geom.shstGeometryId);
                    segmentGeoms.geometry.coordinates.push(segementGeom.geometry.coordinates);
                }
                if (pathCandidate.segments.length > 0)
                    pathCandidate.matchedPath = segmentGeoms;
                return pathCandidate;
            }
            catch (e) {
                return null;
            }
        });
    }
    match(feature) {
        return __awaiter(this, void 0, void 0, function* () {
            //var options = {};
            var pathCandidates = [];
            var bestPathCandidate = null;
            this.pointMatcher.searchRadius = this.searchRadius * 2;
            // do default shst edge look up first
            var startPoint = turfHelpers.point(feature.geometry.coordinates[0]);
            var endPoint = turfHelpers.point(feature.geometry.coordinates[feature.geometry.coordinates.length - 1]);
            var startCandidates = yield this.pointMatcher.getPointCandidates(startPoint, null, 10);
            var endCandidates = yield this.pointMatcher.getPointCandidates(endPoint, null, 10);
            for (var i = 0; i < startCandidates.length; i++) {
                for (var j = 0; j < endCandidates.length; j++) {
                    // if start and end are on the same graph edge make sure they're the same referenceId/correct direction
                    if (startCandidates[i].geometryId == endCandidates[j].geometryId) {
                        // skip if not on the same reference (oposing directions)
                        if (startCandidates[i].referenceId != endCandidates[j].referenceId)
                            continue;
                        // if start location is after end location skip 
                        if (startCandidates[i].location > endCandidates[j].location)
                            continue;
                        var edgeGeom = this.tileIndex.objectIndex.get(startCandidates[i].geometryId);
                        var pathCandidate = new PathCandidate();
                        var matchWorked = true;
                        pathCandidate.score = util_1.rmse([startCandidates[i].calcScore(), endCandidates[j].calcScore()]);
                        pathCandidate.originalFeature = feature;
                        pathCandidate.matchType = MatchType.DIRECT;
                        pathCandidate.segments = [];
                        try {
                            pathCandidate.matchedPath = yield this.tileIndex.geom(startCandidates[i].referenceId, startCandidates[i].location, endCandidates[j].location);
                        }
                        catch (err) {
                            pathCandidate.matchedPath = { type: "Feature", properties: [], geometry: { "type": "LineString", coordinates: [startCandidates[i].pointOnLine.geometry.coordinates[0], startCandidates[i].pointOnLine.geometry.coordinates[1]] } }; //await this.tileIndex.geom(startCandidates[i].referenceId, startCandidates[i].location,  endCandidates[j].location);
                        }
                        var pathSegment = new PathSegment();
                        pathSegment.geometryId = startCandidates[i].geometryId;
                        pathSegment.referenceId = startCandidates[i].referenceId;
                        pathSegment.section = [startCandidates[i].location, endCandidates[j].location];
                        pathSegment.toIntersectionId = edgeGeom.toIntersectionId;
                        var ref = this.tileIndex.objectIndex.get(pathSegment.referenceId);
                        pathSegment.referenceLength = getReferenceLength(ref);
                        pathSegment.fromIntersectionId = edgeGeom.fromIntersectionId;
                        pathCandidate.segments.push(pathSegment);
                        var startDist = distance_1.default(startPoint, startCandidates[i].pointOnLine, { "units": "meters" });
                        var endDist = distance_1.default(endPoint, endCandidates[j].pointOnLine, { "units": "meters" });
                        pathCandidate.score = util_1.rmse([startDist, endDist, pathCandidate.getLengthDelta()]);
                        //console.log( pathCandidate.score)
                        pathCandidates.push(pathCandidate);
                    }
                }
            }
            if (pathCandidates.length > 0) {
                pathCandidates.sort((a, b) => {
                    return a.score - b.score;
                });
                bestPathCandidate = pathCandidates[0];
            }
            if (!bestPathCandidate) {
                // fall back to hmm for probabilistic path discovery
                if (!this.osrm)
                    throw "Graph not buit. call buildGraph() before running queries.";
                var hmmOptions = {
                    coordinates: feature.geometry.coordinates,
                    annotations: true,
                    geometries: 'geojson',
                    radiuses: Array(feature.geometry.coordinates.length).fill(this.searchRadius)
                };
                try {
                    var matches = yield new Promise((resolve, reject) => {
                        this.osrm.match(hmmOptions, function (err, response) {
                            if (err)
                                reject(err);
                            else
                                resolve(response);
                        });
                    });
                    var graphEdgeSequence = [];
                    if (matches['matchings'] && matches['matchings'].length > 0) {
                        var match = matches['matchings'][0];
                        if (MIN_CONFIDENCE < match.confidence) {
                            // this is kind of convoluted due to the sparse info returned in the OSRM annotations...
                            // write out sequence of nodes and edges as emitted from walking OSRM-returned nodes
                            // finding the actual posistion and directionality of the OSRM-edge within the ShSt graph 
                            // edge means that we have to snap start/end points in the OSRM geom
                            //console.log(JSON.stringify(match.geometry));
                            var edgeCandidates;
                            var nodes = [];
                            var visitedNodes = new Set();
                            // ooof this is brutual -- need to unpack legs and reduce list... 
                            for (var leg of match['legs']) {
                                //console.log(leg['annotation']['nodes'])
                                for (var n of leg['annotation']['nodes']) {
                                    if (!visitedNodes.has(n) || nodes.length == 0)
                                        nodes.push(n);
                                    visitedNodes.add(n);
                                }
                            }
                            var visitedEdges = new Set();
                            // then group node pairs into unique edges...
                            for (var nodeId of nodes) {
                                if (yield this.db.has('node:' + nodeId)) {
                                    var node = JSON.parse(yield this.db.get('node:' + nodeId));
                                    var nodeEdges = JSON.parse(yield this.db.get('node-edges:' + nodeId));
                                    if (!edgeCandidates) {
                                        edgeCandidates = new Set([...nodeEdges.edgeIds]);
                                    }
                                    else {
                                        var newCandiates = new Set([...nodeEdges.edgeIds]);
                                        edgeCandidates = new Set([...edgeCandidates].filter(x => newCandiates.has(x)));
                                    }
                                    if (edgeCandidates.size == 1) {
                                        var edgeId = [...edgeCandidates][0];
                                        // reduce repeating edgeIds down to one per edge
                                        if (graphEdgeSequence.length == 0 || graphEdgeSequence[graphEdgeSequence.length - 1].edgeId !== edgeId) {
                                            var edge = JSON.parse(yield this.db.get('edge:' + edgeId));
                                            graphEdgeSequence.push(edge);
                                        }
                                        edgeCandidates = new Set([...nodeEdges.edgeIds]);
                                    }
                                }
                            }
                        }
                    }
                    var startPoint = turfHelpers.point(match.geometry.coordinates[0]);
                    var endPoint = turfHelpers.point(match.geometry.coordinates[match.geometry.coordinates.length - 1]);
                    var startCandidates = yield this.pointMatcher.getPointCandidates(startPoint, null, 10);
                    var endCandidates = yield this.pointMatcher.getPointCandidates(endPoint, null, 10);
                    startCandidates = startCandidates.filter((item) => {
                        return item.geometryId === graphEdgeSequence[0].shstGeometryId;
                    });
                    endCandidates = endCandidates.filter((item) => {
                        return item.geometryId === graphEdgeSequence[graphEdgeSequence.length - 1].shstGeometryId;
                    });
                    var alreadyIncludedPaths = new Set();
                    for (var i = 0; i < startCandidates.length; i++) {
                        for (var j = 0; j < endCandidates.length; j++) {
                            // if start and end are on the same graph edge make sure they're the same referenceId/correct direction
                            if (startCandidates[i].geometryId == endCandidates[j].geometryId) {
                                // skip if not on the same reference (oposing directions)
                                if (startCandidates[i].referenceId != endCandidates[j].referenceId)
                                    continue;
                                // if start location is after end location skip 
                                if (startCandidates[i].location > endCandidates[j].location)
                                    continue;
                            }
                            var pathCandidate = new PathCandidate();
                            var matchWorked = true;
                            pathCandidate.matchType = MatchType.HMM;
                            pathCandidate.score = match.confidence;
                            pathCandidate.originalFeature = feature;
                            pathCandidate.matchedPath = turfHelpers.feature(match.geometry);
                            //console.log(JSON.stringify(pathCandidate.matchedPath));
                            pathCandidate.segments = [];
                            var length = pathCandidate.getOriginalFeatureLength();
                            for (var k = 0; k < graphEdgeSequence.length; k++) {
                                if ((k == 0 && startCandidates[i].geometryId !== graphEdgeSequence[k].shstGeometryId) &&
                                    (k == graphEdgeSequence.length - 1 && endCandidates[j].geometryId != graphEdgeSequence[k].shstGeometryId)) {
                                    matchWorked = false;
                                    break;
                                }
                                var pathSegment = new PathSegment();
                                pathSegment.geometryId = graphEdgeSequence[k].shstGeometryId;
                                pathCandidate.segments.push(pathSegment);
                            }
                            if (pathCandidate.segments.length > 0) {
                                pathCandidate.startPoint = startCandidates[i];
                                // build directionality into edge sequences   
                                for (var k = 0; k < pathCandidate.segments.length; k++) {
                                    var edgeGeom = this.tileIndex.objectIndex.get(pathCandidate.segments[k].geometryId);
                                    // if start and end are on the same graph edge make sure they're the same referenceId/direction
                                    if (startCandidates[i].geometryId == endCandidates[j].geometryId) {
                                        pathCandidate.endPoint = endCandidates[j];
                                        pathCandidate.segments[k].referenceId = pathCandidate.startPoint.referenceId;
                                        pathCandidate.segments[k].section = [startCandidates[i].location, endCandidates[j].location];
                                        var ref = this.tileIndex.objectIndex.get(pathCandidate.segments[k].referenceId);
                                        pathCandidate.segments[k].referenceLength = getReferenceLength(ref);
                                    }
                                    else {
                                        if (k == 0)
                                            pathCandidate.segments[k].referenceId = pathCandidate.startPoint.referenceId;
                                        else {
                                            if (pathCandidate.segments[k - 1].toIntersectionId == edgeGeom.fromIntersectionId) {
                                                pathCandidate.segments[k].referenceId = edgeGeom.forwardReferenceId;
                                            }
                                            else if (pathCandidate.segments[k - 1].toIntersectionId == edgeGeom.toIntersectionId && edgeGeom.backReferenceId)
                                                pathCandidate.segments[k].referenceId = edgeGeom.backReferenceId;
                                            else {
                                                matchWorked = false;
                                                break;
                                            }
                                        }
                                        var ref = this.tileIndex.objectIndex.get(pathCandidate.segments[k].referenceId);
                                        pathCandidate.segments[k].referenceLength = getReferenceLength(ref);
                                        if (k == pathCandidate.segments.length - 1) {
                                            pathCandidate.endPoint = endCandidates[j];
                                            pathCandidate.segments[k].section = [0, endCandidates[j].location];
                                        }
                                        else if (k == 0) {
                                            pathCandidate.segments[k].section = [pathCandidate.startPoint.location, pathCandidate.segments[k].referenceLength];
                                        }
                                        else {
                                            pathCandidate.segments[k].section = [0, pathCandidate.segments[k].referenceLength];
                                        }
                                    }
                                    // put to/from on semgnet
                                    if (edgeGeom.forwardReferenceId == pathCandidate.segments[k].referenceId) {
                                        pathCandidate.segments[k].fromIntersectionId = edgeGeom.fromIntersectionId;
                                        pathCandidate.segments[k].toIntersectionId = edgeGeom.toIntersectionId;
                                    }
                                    else {
                                        // reverse to/from for back references
                                        pathCandidate.segments[k].fromIntersectionId = edgeGeom.toIntersectionId;
                                        pathCandidate.segments[k].toIntersectionId = edgeGeom.fromIntersectionId;
                                    }
                                }
                                var refIdHash = uuidHash(pathCandidate.segments.map((value) => { return value.referenceId; }).join(' '));
                                if (matchWorked && !alreadyIncludedPaths.has(refIdHash)) {
                                    alreadyIncludedPaths.add(refIdHash);
                                    bestPathCandidate = pathCandidate;
                                }
                            }
                        }
                    }
                }
                catch (e) {
                    // no-op failed to match
                }
            }
            if (bestPathCandidate) {
                var cleanedPath = [];
                var segCoords = [];
                for (var i = 0; i < bestPathCandidate.segments.length; i++) {
                    var segment = bestPathCandidate.segments[i];
                    // adding fudge factor for decimal precision issues
                    if (segment.section[0] < segment.section[1] + this.searchRadius && segment.section[1] <= segment.referenceLength + this.searchRadius && segment.section[0] + this.searchRadius >= 0) {
                        if (this.snapIntersections && (segment.section[1] - segment.section[0]) > this.searchRadius) {
                            if (i == 0 && segment.referenceLength - segment.section[0] < this.searchRadius)
                                continue;
                            if (i == 0 && segment.section[0] < this.searchRadius)
                                segment.section[0] = 0;
                            if (i == bestPathCandidate.segments.length - 1 && segment.section[1] < this.searchRadius)
                                continue;
                            if (i == bestPathCandidate.segments.length - 1 && segment.referenceLength - segment.section[1] < this.searchRadius)
                                segment.section[1] = segment.referenceLength;
                            if (i > 0 && i < bestPathCandidate.segments.length - 1) {
                                segment.section[0] = 0;
                                segment.section[1] = segment.referenceLength;
                            }
                        }
                        if (segment.section[0] == segment.section[1])
                            continue;
                        var segGeom = yield this.tileIndex.geom(segment.referenceId, segment.section[0], segment.section[1]);
                        if (segGeom) {
                            cleanedPath.push(segment);
                            segCoords.push(segGeom.geometry.coordinates);
                        }
                    }
                }
                bestPathCandidate.segments = cleanedPath;
                if (cleanedPath.length == 0)
                    return null;
                if (segCoords.length > 0) {
                    var segmentGeoms = turfHelpers.multiLineString([]);
                    segmentGeoms.geometry.coordinates = [...segCoords];
                    bestPathCandidate.matchedPath = segmentGeoms;
                }
            }
            return bestPathCandidate;
        });
    }
}
exports.Graph = Graph;
class PathSegment {
}
exports.PathSegment = PathSegment;
class PathCandidate {
    toDebugView() {
        var debugCollection = turfHelpers.featureCollection([this.originalFeature, this.matchedPath]);
        return debugCollection;
    }
    getOriginalFeatureLength() {
        if (!this.originalFeatureLength)
            this.originalFeatureLength = length_1.default(this.originalFeature, { "units": "meters" });
        return this.originalFeatureLength;
    }
    getPathLength() {
        if (!this.pathLength) {
            this.pathLength = 0;
            for (var segment of this.segments) {
                if (segment.section)
                    this.pathLength = this.pathLength + (segment.section[1] - segment.section[0]);
                else
                    this.pathLength = this.pathLength + segment.referenceLength;
            }
        }
        return this.pathLength;
    }
    getLengthDelta() {
        return this.getPathLength() - this.getOriginalFeatureLength();
    }
    isColinear(candidate) {
        if (this.segments.length > 0 && candidate.segments.length > 0 && this.segments.length == candidate.segments.length) {
            var path1GeometryIds = new Set();
            var path2GeometryIds = new Set();
            for (var segment of this.segments) {
                path1GeometryIds.add(segment.geometryId);
            }
            for (var segment of candidate.segments) {
                path2GeometryIds.add(segment.geometryId);
                if (!path1GeometryIds.has(segment.geometryId))
                    return false;
            }
            for (var segment of this.segments) {
                if (!path2GeometryIds.has(segment.geometryId))
                    return false;
            }
        }
        else
            return false;
        return true;
    }
}
exports.PathCandidate = PathCandidate;
